from typing import Dict
from unittest import IsolatedAsyncioTestCase, mock

from easybits_module_interface.models import Message, MessageTypes


class BaseModuleTestCase(IsolatedAsyncioTestCase):
    """
    Base class for module tests.

    Inherit from this class to perform tests onto your module's callback method.

    Example:
        Suppose you've developed a module that echos the users input.
        The following snippet will test two cases

        #. success: An incoming message is provided, we echo the user input
        #. failure: No incoming message is provided, we respond with a default error message

            >>> class EchoModuleTestCase(BaseModuleTestCase):
            ...     def setUp(self):
            ...         super().setUp()
            ...         self.instance = EchoModule()
            ...
            ...     async def test_success(self):
            ...         await self.perform_test({'message': {'incoming': 'hello world!'}}, 'hello world')
            ...
            ...     async def test_failure(self):
            ...         await self.perform_test({'message': {'incoming': ''}}, 'Something went wrong.')
    """
    def setUp(self):
        super().setUp()

        pika_mock = mock.patch('easybits_module_interface._internal.aoi_pika', mock.AsyncMock(), create=True)
        self.addCleanup(pika_mock.stop)
        self.pika_mock = pika_mock.start()

        pub_entity_mock = mock.Mock(return_value=mock.Mock(publish=mock.AsyncMock()))
        pub_mock = mock.patch('easybits_module_interface._internal.Publisher', pub_entity_mock)
        self.addCleanup(pub_mock.stop)
        self.pub_mock = pub_mock.start()

    @property
    def default_message(self) -> Dict:
        """
        Property of default message dictionary.

        Hint:
            You can override this property to provide a custom default message.

        :returns: default message dictionary
        """
        return {
            "meta": {
                "type": MessageTypes.INCOMING_MESSAGE.value,
                "botId": 1,
                "event": None,
                "messageId": None,
                "controller": None,
                "internalUserId": 1,
                "communicationChannelId": 1
            },
            "config": None,
            "message": None,
            "context": None,
        }

    def _build_message(self, message: Dict, message_type: MessageTypes = MessageTypes.INCOMING_MESSAGE) -> Dict:
        """
        Extends default content to a given message.

        :param message: Message dictionary
        :returns: extended message containing default ``meta`` data
        """
        default_message = self.default_message.copy()
        default_message['meta']['type'] = message_type.value
        if 'meta' in message:
            default_message['meta'].update(message.pop('meta'))
        if 'config' in message and default_message['config'] is not None:
            default_message['config'].update(message.pop('config'))
        if 'message' in message and default_message['message'] is not None:
            default_message['message'].update(message.pop('message'))
        if 'context' in message and default_message['context'] is not None:
            default_message['context'].update(message.pop('context'))

        default_message.update(message)
        return default_message

    async def perform_test(self, message: Dict, expected_message_content: str) -> None:
        """
        Helper method to perform a test on the ``callback`` method of a module implementation.

        :param message: a message dictionary containing the required atttributes
        :param expected_message_content: The expected content of ``message.outgoing``
        :returns: None
        """
        assert self.instance is not None, 'You must assign a module to self.instance.'
        await self.instance.callback(Message.from_dict(**self._build_message(message)))

        self.pub_mock().publish.assert_called_once()
        self.assertEqual(
            self.pub_mock().publish.call_args_list[0][0][0].message.outgoing,
            expected_message_content
        )

    async def perform_responded_test(self):
        """Test case to test whether a module processes relevant messages."""
        assert self.instance is not None, 'You must assign a module to self.instance.'
        message = self._build_message({'message': {'incoming': 'Hi!'}})
        await self.instance.callback(Message.from_dict(**self._build_message(message)))

        # at least 1 call to publish
        self.assertGreaterEqual(self.pub_mock().publish.call_count, 1)

    async def perform_ignored_test(self):
        """Test case to test whether a module ignores irrelevant messages."""
        assert self.instance is not None, 'You must assign a module to self.instance.'
        message = self._build_message({}, MessageTypes.OUTGOING_MESSAGE)
        await self.instance.callback(Message.from_dict(**self._build_message(message)))

        self.pub_mock().publish.assert_not_called()

    async def test_compatibility(self):
        """Test to validate compatibility of module implementation."""
        if self.__class__ is BaseModuleTestCase:
            self.skipTest('Not required for base class.')
        self.pub_mock().publish.reset_mock()
        await self.perform_ignored_test()
        self.pub_mock().publish.reset_mock()
        await self.perform_responded_test()



from concurrent.futures import ThreadPoolExecutor
from uuid import uuid4


class ManagedThreadPoolExecutor:
    """
    Implementation of a managed ThreadPoolExecutor.

    Allows seamless scheduling of unstable function calls.
    This implementation will spin up new instances of ThreadPoolExecutor
    if threads are supposed to be cancelled.

    See usage inside `./thread_executor.py`
    """
    def __init__(self, max_workers=None, logger=None):
        self.max_workers = max_workers
        self.current_executor = None
        self.current_executor_id = None
        self.frozen_executors = {}
        self.task_map = {}
        self.frozen_tasks = {}
        self.logger = logger
        self.requires_restart = False

    def _get_executor(self):
        """
        Factory function for new ThreadPoolExecutor instances

        :returns: Tuple of ThreadPoolExecutor instance and it's UID
        """
        return ThreadPoolExecutor(max_workers=self.max_workers), str(uuid4())

    def submit(self, func, process_id, *args, **kwargs):
        """
        Method to submit async Callable with its ID to the current thread pool executor.

        :param func: An async Callable to execute
        :param process_id: UID of the process
        :returns: A Future for the callable
        """
        if self.current_executor is None:
            self.logger.debug('Creating new executor')
            self.current_executor, self.current_executor_id = self._get_executor()
        if self.current_executor_id not in self.task_map:
            self.task_map[self.current_executor_id] = []

        self.task_map[self.current_executor_id].append(process_id)

        return self.current_executor.submit(func, *args, **kwargs)

    def shutdown(self) -> None:
        """
        Method to shut down all executors in the running ManagedThreadPoolExecutor instance.

        :returns: Nothing
        """
        if self.current_executor is not None:
            self.current_executor.shutdown()
        for executor in self.frozen_executors:
            executor.shutdown(wait=False)

        self.frozen_executors.clear()
        self.current_executor = None
        self.current_executor_id = None

    def _remove_executor(self, executor_id) -> None:
        """
        Helper method to _gracefully_ remove an executor.

        :param executor_id: the ID of the executor
        :returns: Nothing
        """
        if executor_id in self.frozen_executors:
            self.frozen_executors[executor_id].shutdown(wait=False)
        self.frozen_executors = {k: v for k, v in self.frozen_executors.items() if k != executor_id}
        self.task_map = {k: v for k, v in self.task_map.items() if k != executor_id}
        self.frozen_tasks = {k: v for k, v in self.frozen_tasks.items() if k != executor_id}

    def _handle_task_list(self, task_list, is_cancelled=False):
        """
        Method to handle finished tasks.

        This method "freezes" thread pools and resets the current executor if it contains
        a stale task.

        :param task_list: List of finished tasks
        :param is_cancelled: Indicates that `task_list` is a list of cancelled tasks
        :returns: A list of executor IDs of executors that can be safely shut down.
        """
        executors_to_remove = []
        for idx in task_list:
            for executor_id, tasks in filter(lambda x: idx in x[1], self.task_map.items()):
                if is_cancelled:
                    if executor_id == self.current_executor_id:
                        self.frozen_executors[executor_id] = self.current_executor
                        self.current_executor = None
                        self.current_executor_id = None
                        self.requires_restart = True

                    if executor_id not in self.frozen_tasks:
                        self.frozen_tasks[executor_id] = []
                    self.frozen_tasks[executor_id].append(idx)

                self.task_map[executor_id] = [i for i in self.task_map[executor_id] if i != idx]
                # if no more unfrozen tasks for frozen executor, remove it
                if len(self.task_map[executor_id]) == 0 and executor_id != self.current_executor_id:
                    executors_to_remove.append(executor_id)
        return executors_to_remove

    def garbage_collect(self, finished_tasks, cancelled_tasks) ->None:
        """
        Method to garbage collect inside of running thread pools.

        :param finished_tasks: List of tasks that are finished
        :param cancelled_tasks: List of tasks that are cancelled
        :returns: Nothing
        """
        # remove finished tasks from task_map
        executors_to_remove = (
            self._handle_task_list(finished_tasks)
            + self._handle_task_list(cancelled_tasks, is_cancelled=True)
        )

        for executor_id, executors in self.frozen_executors.items():
            if self.task_map.get(executor_id, None) in [None, []]:
                executors_to_remove.append(executor_id)

        for executor_id in set(executors_to_remove):
            self.logger.debug(f'Removing executor {executor_id}')
            self._remove_executor(executor_id)





import asyncio
from dataclasses import dataclass, field
from datetime import datetime
from itertools import count
from typing import Callable, Optional

from easybits_module_interface.models import Message


@dataclass
class GenerationRequest:
    """
    Dataclass to hold the request for a generation task.
    """
    message: Message
    method: Callable
    verbose: int
    timeout: int
    owner_id: str = field(default_factory=lambda: 'unknown')
    timestamp: datetime = field(default_factory=datetime.now)
    notified: bool = False
    task_id: int = field(default_factory=count().__next__, init=False)
    # timeout in minutes
    execution_timestamp: Optional[datetime] = None
    execution_timeout: int | float = 30

    def __post_init__(self):
        if self.verbose == 0:
            self.notified = True


@dataclass
class RunningTask:
    """
    Dataclass to hold the running tasks.
    """
    generation_request: GenerationRequest
    task: asyncio.Future

